#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

extern "C"
{
  void bcopy(char*, char*, int) ;
  void bzero(char*, int) ;
}

const	int	TREE[4] = { 0x1, 0x2, 0x4, 0x8 } ;

// ----------------------------------------------------------------------------------------
// ComputesNodesPerPage - given an assignment of trees to pages and the number of nodes in
//			  in the trees nodes[]computes the number of
//			  nodes in that will go in each of the page. 
// ----------------------------------------------------------------------------------------
void
ComputeNodesPerPage( int assign[], int treeNodes[], int nodes[] )
{
	// For each of page 
	for ( int q = 0 ; q < 4 ; q++ )
	{
		for ( int tree = 0 ; tree < 4 ; tree++ )
		{
			if ( assign[q] & (1 << tree) )	
				nodes[q] += treeNodes[tree] ; 
		}
	}
}
	
	
double
Penalty( int treeNodes[], int size, int page[], int pageCount ) 
{
	assert( pageCount != 0 ) ;

	double m = double(size/pageCount) ;     	

	int n[4] ; 

	bzero( (char *)n,sizeof(n) ) ;

	ComputeNodesPerPage(page,treeNodes,n) ; 
			
	double sum = 0.0 ;	

	for ( int q = 0 ; q < 4 ; q++ )
	{
		sum += (n[q] - m) * (n[q] - m) ;
	}

	return sum ; 
}

// ----------------------------------------------------------------------------------------
// CheckValidAssignment - given an assignment of trees to pages and the bucket occupancy 
//			  checks whether the given assignment is valid OR not.
//				
//			  returns 0 if valid
//			  else 1 
// ----------------------------------------------------------------------------------------

int
CheckValidAssignment( int assign[], int treeNodes[], int M )
{
	int n[4] = {0,0,0,0} ;

	ComputeNodesPerPage(assign,treeNodes,n) ;

	for ( int q = 0 ; q < 4 ; q++ )
	{
		if ( n[q] > M )
			return 1 ; 
	}

	return 0 ;
}
		
// ----------------------------------------------------------------------------------------
// Partition - Given the number of nodes in each of the subtree of a node in 
//	       quadtree (treeNodes) and bucket occupancy (M)		
//	       returns the optimal assignments of these subtrees to pages. 
//				
// treeAssignments returns 4 integers. Each of them contain a value between 0 thru 15     
// 
// Only last 4 bits of each integer is used to find the tree assignment.
// Assignments are encoded as follows :
//
//      bit3    bit2    bit1    bit0 
//	 |	 |	 |	 |
//	 |	 |	 | 	 |
//	 v	 v	 v 	 v	
//      tree3   tree2   tree1   tree0
//
// If bit3 is set in treeAssignments[1] then tree3 is assigned to page1
// If treeAssignments[0] is zero then the page is empty and no tree has been assigned
// to it.
// The nodes in the subtree should be less than bucket occupancy that can be
// accomodated in a page.
// ----------------------------------------------------------------------------------------

int
Partition( int treeNodes[], int treeAssignments[], int M )  
{
	int i,j,k,l ; 

	int pageCount = 0 ; 

	int bestPageCount = 10000 ; 

	int page[4] ; 

	int total = 0 ; 

	// Compute the total number of nodes in all the 
	// given subtrees
	for ( i = 0 ; i < 4 ; total += treeNodes[i++] ) ; 
	
	bzero( (char *)treeAssignments,sizeof(int) * 4 ) ;

	bzero( (char *)page,sizeof(page) ) ; 	

	// Enumerate all combinations for assignment
	// of trees to pages

	// Tree 1
	for ( i = 0 ; i < 4 ; i++ )
	{
		// Tree 2
		for ( j = 0 ; j < 4 ; j++ )
		{
			// Tree 3
			for ( k = 0 ; k < 4 ; k++ )
			{
				// Tree 4
				for ( l = 0 ; l < 4 ; l++ )
				{
					// Encode the assignment in page arrays
					// Assign Tree 1 to the page
					page[i] |= TREE[0] ;

					// Assign Tree 2 to the page
					page[j] |= TREE[1] ;

					// Assign Tree 3 to the page
					page[k] |= TREE[2] ;

					// Assign Tree 4 to the page
					page[l] |= TREE[3] ;

					// Check the assignment for each page is less than
					// the required bucket occupancy
					// If not reject the combination as invalid

					if ( !CheckValidAssignment(page,treeNodes,M) )
					{
						// Count the number of pages for this assignment
						for ( int q = 0 ; q < 4 ; 
							page[q++] ? pageCount++ : pageCount ) ;
					
						if ( pageCount < bestPageCount )
						{
							bestPageCount = pageCount ; 	
	
							bcopy( (char *)page, (char *)treeAssignments, sizeof(page) ) ; 	

						}

						if ( pageCount == bestPageCount )
						{
							// Compute Penalty function
							double bestPen = Penalty(treeNodes,total,treeAssignments,bestPageCount) ;

							double currentPen = Penalty(treeNodes,total,page,pageCount) ; 
		
							assert ( bestPen >= 0 && currentPen >= 0 ) ; 

							if ( currentPen < bestPen )
							{
								bcopy( (char *)page, (char *)treeAssignments, sizeof(page) ) ; 	
							}

						}
								
					}
					
					bzero( (char *)page,sizeof(page) ) ; 	

					pageCount = 0 ;

				} // end for tree 4
			} // end for tree 3
		} // end for tree 2
	} // end for tree 1 

				
	return bestPageCount ;	

}	

void
sort( int path[], int out[] ) 
{
  int localcopy[4] ;

  for ( int i = 0 ; i < 4 ; localcopy[i] = path[i++] ) ;
  
  for ( i = 0 ; i < 4 ; i++ )
  {
    int max = 0 ;
    for ( int j = 0 ; j < 4 ; 
	  max = localcopy[max] < localcopy[j] ? j : max, j++ ) ; 
    out[3-i] = max ;
    localcopy[max] = -2 ;  
  }
}


			
/*
void
main()
{
	int t[4] ;

	int n[4] ;

	int i ;

	for ( i = 0 ; i < 4 ; i++ )
	{
		printf("Enter tree size %d :", i ) ;

		scanf("%d",&t[i] ) ;
	}

	printf("Required Pages: %d\n", Partition(t,n,250) ) ;

	printf("Best Assignment %d %d %d %d\n",n[0],n[1],
				n[2],n[3]) ;  

	printf("%d %d %d %d\n",t[0],t[1],
				t[2],t[3]) ;  
	sort(t,n) ;

	printf("Sort %d %d %d %d\n",t[n[0]],t[n[1]],
				t[n[2]],t[n[3]]) ;  
}
*/
