#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <iostream.h>
#include <math.h>
#include <fcntl.h>
#include <assert.h>

#include "standard.h"
#include "qtree.h"
#include "quadsort.h"



Quadfile_t *file;
Point_t *points;

int mycompare(const void *p1, const void *p2)
{
 return compare((Point_t*)p1, (Point_t*)p2);
}

/*
 * returns 1 if morton(p1) >= morton(p2), 0 otherwise 
 */
int compare(const Point_t *p1, const Point_t *p2)
{
  unsigned int dx = (unsigned)p1->x ^ (unsigned)p2->x;
  unsigned int dy = (unsigned)p1->y ^ (unsigned)p2->y;
  unsigned int d  = dx ^ dy;
  return ((d<dx && d<dy) || dy>dx) ? p1->y >= p2->y : p1->x >= p2->x;
}


Quadfile_t::   
Quadfile_t(int fd, Point_t pt1, Point_t pt2)
{

  fDesc = fd;
  first = (Point_t*)malloc(sizeof(Point_t));
  CHECK_MALLOC(first);
  second = (Point_t*)malloc(sizeof(Point_t));
  CHECK_MALLOC(second);
  
  first->x = pt1.x;
  first->y = pt1.y;
  second->x = pt2.x;
  second->y = pt2.y;
}

Quadfile_t::   
~Quadfile_t()
{
  close(fDesc);
  if (first)
   free(first);
  if (second)
   free(second);
}

/*
 * Checks if the given point lies within the world extent
 */   
int isValid(const Point_t *pt)
{
  if (pt)
    return 1;
  return 0;
}

/* 
 * returns the number of points between the current point
 * (given by first) and pt BOTH INCLUDED
 * Return Values:
 *   
 * 
 *
 */
int Quadfile_t::   
peek(const Point_t &pt)
{
  if (!isValid(&pt))	 	/* Invalid pt , cannot peek */
    return -1;			
  if (!isValid(first))		/* possible end of file     */
  {
     assert(!isValid(second));	/* second better be invalid too */
     return -1;
  }
  if (!compare(&pt,first))  			   /* pt < first, return 0 */
    return 0;
  if (compare(&pt, first) && compare(first, &pt))    /* pt == first,return 1 */
    return 1;
  /* pt > first is true */
  assert(compare(&pt, first) && !compare(first, &pt));
  if (!isValid(second))
   return 1;
  if (!compare(second, &pt))			    /* pt < second  */
   return 1;
  if (compare(&pt,second) && compare(second,&pt)) /* pt == second */
   return 2;
  return 2;					   /* pt > second, return 2 */
}     

/*
 * moves the current file pointer to the first point in the file that
 * is greater than pt (if one exists)
 * if EOF is reached before such a point can be found, file pointer
 * is set at EOF, and buffers invalidated
 * The file pointer is moved by 2 extra positions to reset the buffers
 * and then moved back
 * Return Values:
 *   0 if the file ptr is at the very end of the file (first and second are INVALID
 *   1 otherwise
 */
int Quadfile_t::   
move(const Point_t &pt)
{
int nbytes;
Point_t cur_pt;

  if (!isValid(first))
  {
     assert(!isValid(second));  /* second better be invalid too */
     return 0;
  }
  if (compare(first,&pt) && !compare(&pt,first))   /* first > pt, cannot move */
    return 1;
  /* first <= pt , in either case I need to move*/
  if (!isValid(second)) /* invalidate first. fDesc should be at end of file */
  {
    nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t));
    assert(nbytes == sizeof(Point_t));  /* first was valid */
    first = NULL;			/* first no longer valid */
    return 0;
  }
  /* both first and second valid && first <= pt */
  if (compare(second, &pt) && !compare(&pt, second)) /* second > pt */
  {
    nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t));
    assert(nbytes == sizeof(Point_t));  /* first was valid */
    nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t));
    assert(nbytes == sizeof(Point_t));  /* second was valid too */
    
    /* peek at two more record to reset values of first and second */
    nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t));  
    if (nbytes != sizeof(Point_t)) /* no more records available */
    {
      first = NULL;
      second = NULL;
      return 0;
    }
    first->x = cur_pt.x;
    first->y = cur_pt.y;
    nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t));  
    if (nbytes != sizeof(Point_t)) /* no more records available */
    { 
     second = NULL;
     UNGET_ONEREC(fDesc);
     return 1;
    }
    second->x = cur_pt.x;
    second->y = cur_pt.y;
    UNGET_TWORECS(fDesc);
    return 1;
  }
  /* second <= pt, search forward until the right position */

  /* read so long as cur_pt <= pt */
  while ( ((nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t))) 
          == sizeof(Point_t)) &&
          (!compare(&cur_pt, &pt) || (compare(&cur_pt,&pt) && compare(&pt, &cur_pt))) );
  if (nbytes != sizeof(Point_t)) /* no more records */
  {
    /* Invalidate first and second */
    first = second = NULL;
    return 0;
  }
  first->x = cur_pt.x;
  first->y = cur_pt.y;
  /* read ahead for one more record */
  nbytes = read(fDesc, (char*)&cur_pt, sizeof(Point_t));
  if (nbytes != sizeof(Point_t))
  {
    UNGET_ONEREC(fDesc);
    second = NULL;
    return 1;
  } 
  second->x = cur_pt.x;
  second->y = cur_pt.y;
  UNGET_TWORECS(fDesc);
  return 1;
}          


int load_sort(char *inFile, char *outFile, int numrecs)
{
int infDesc;			/* bulk load file */
int outfDesc;                   /* sorted file    */
int count; 			/* Number of records processed */
Point_t pt;
int nbytes;

  infDesc = open(inFile, O_RDONLY, 0);
  if (infDesc < 0)
  {
     fprintf(stderr,"Unable to open %s :", inFile);
     perror("");
     return 0;
  }
  outfDesc = open(outFile, O_WRONLY, 0);
  if (outfDesc < 0)
  {
     fprintf(stderr,"Unable to open %s :", outFile);
     perror("");
     close(infDesc);
     return 0;
  }
  count = 0;
  while (count < numrecs)
  {
     nbytes = read(infDesc, (char*)&pt, sizeof(Point_t));
     if (nbytes != sizeof(Point_t))
     {
       fprintf(stderr, "Unexpected end of file:");
       perror("");
       close(infDesc);
       close(outfDesc);
       return 0;
     }
     points[count++] = pt;
  }
  close(infDesc);
  /* done loading. Now sort and store in outfile */
  qsort(points,0, numrecs-1, mycompare);
  file = new Quadfile_t(outfDesc, points[0], points[1]);  
  count = 0;
  while (count < numrecs)
  {
    nbytes = write(outfDesc, (char*)&points[count++], sizeof(Point_t));
    if (nbytes != sizeof(Point_t))
    {
      fprintf(stderr,"Error writing into output file:");
      perror("");
      close(outfDesc);
      return 0;
    }
  }
  close(outfDesc);   
	return 1;
}   

void usage()
{
  fprintf(stderr,"Usage: quadsort <infile> <outfile> <numrecords> \n");
  exit(-1);
}

/*
main(int argc, char **argv)
{
int numrecs = atoi(argv[argc-1]);

  if (argc < 4)
    usage();
  points = (Point_t*)malloc(numrecs*sizeof(Point_t));
  if (!load_sort(argv[argc-2], argv[argc-3], numrecs))
    exit(-1);
  fprintf(stderr,"Sorting, Loading done..\n");
}
*/
